# Graph Search vs. Dynamic Programming vs Greedy Algorithms

*   Idea behind *graph search* design:
    *   Frame the problem as a problem on a graph
    *   Explore vertices and edges using a worklist algorithm
    *   Compute a subproblem result for each vertex
    *   Postprocess to extract a final result

*   Idea behind *dynamic programming* design:
    *   Frame the problem as a problem on a directed acyclic graph (DAG)
    *   Explore vertices in topological order
    *   Compute a subproblem result for each vertex
    *   Postprocess to extract a final result

*   Idea behind *greedy* design:
    *   Frame the problem as a problem on a directed acyclic graph (DAG)
    *   Start at a source vertex
    *   Compute and follow a "best" outgoing edge at each vertex (without exploring anything else)
    *   Postprocess to extract a final result

--------------------------------------------------------------------------------

# Problem 1:

A vehicle can travel up to a distance of 10 without refueling, and along its route, there are nine locations, `a` through `i`, where it can be refueled; these locations are spaced as follows:

*   The distance from `a` to `b` is 5.
*   The distance from `b` to `c` is 3.
*   The distance from `c` to `d` is 3.
*   The distance from `d` to `e` is 8.
*   The distance from `e` to `f` is 3.
*   The distance from `f` to `g` is 2.
*   The distance from `g` to `h` is 2.
*   The distance from `h` to `i` is 4.

If the vehicle is to start empty at `a` and end up full at `i`, where should it refuel to minimize the number of refuelings?

Example: The vehicle could:

*   Fuel at `a`,
*   Travel 5 to fuel at `b`,
*   Travel 6 to fuel at `d`,
*   Travel 8 to fuel at `e`,
*   Travel 3 to fuel at `f`,
*   Travel 4 to fuel at `h`, and
*   Travel 4 for fuel at `i`.

This plan takes seven refuelings, but there are plans with fewer.

--------------------------------------------------------------------------------

# Dynamic Programming for the Fueling Problem

Problem: Given a sequence of fueling sites, the distances between consecutive sites, and a maximum range per fueling, find a plan for traveling over the sequence with as few refuelings as possible.

## DAG

*   Vertices (situations):
    *   Locations
*   Edges (actions):
    *   Ways to advance to a location up to `range` units of distance away
*   Edge weights (optional):
    *   Distance to the next location
*   Topological order:
    *   Order along the route: `a → b → … → i`
*   Goal:
    *   Find the path from `a` to `i` with the fewest edges

## Backpointer Class

*   Information for the final result:
    *   The previous location
*   Information to go back:
    *   Nothing additional (since we already have the previous location)
*   Information to compare quality:
    *   The number of fuelings so far

## Choosing a Backpointer

*   Exhaustive search:
    *   Generate all preceding locations out to a distance of `range`
    *   Check that the backpointer minimizes the number of fuelings so far

## Example

*   The range per fueling is 10.
*   The distance from `a` to `b` is 5.
*   The distance from `b` to `c` is 3.
*   The distance from `c` to `d` is 3.
*   The distance from `d` to `e` is 8.
*   The distance from `e` to `f` is 3.
*   The distance from `f` to `g` is 2.
*   The distance from `g` to `h` is 2.
*   The distance from `h` to `i` is 4.

    Location  Previous Location  Fuelings so Far
    --------  -----------------  ---------------
    a         ⊥                  1
    b         a                  2
    c         a                  2
    d         c                  3
    e         d                  4
    f         e                  5
    g         e                  5
    h         e                  5
    i         h                  6

    Reversed Path
    ----
    i → h → e → d → c → a

--------------------------------------------------------------------------------

# Greedy Algorithm for the Fueling Problem

Problem: [same as above]

## DAG

*   [same as above]

## Choosing a Forward Edge

*   Exhaustive search:
    *   Generate …
    *   Check that …

## Example

*   The range per fueling is 10.
*   The distance from `a` to `b` is 5.
*   The distance from `b` to `c` is 3.
*   The distance from `c` to `d` is 3.
*   The distance from `d` to `e` is 8.
*   The distance from `e` to `f` is 3.
*   The distance from `f` to `g` is 2.
*   The distance from `g` to `h` is 2.
*   The distance from `h` to `i` is 4.

    Location  Next Location
    --------  -------------
    a         …
    …

    Path
    ----
    a → …

--------------------------------------------------------------------------------

# Problem 2:

You are creating a lossless compression scheme for text that uses the symbols 'a' through 'd' at the following rates:

*   'a' occurs about 15% of the time in a typical text.
*   'b' occurs about 30% of the time in a typical text.
*   'c' occurs about 25% of the time in a typical text.
*   'd' occurs about 30% of the time in a typical text.

The compressed text will be saved in a binary format, so your compression scheme needs to encode each symbol into a sequence of zeros and ones.  For it to be lossless, there can be no ambiguities in the encoding.  For example, the encoding

*   'a' → `0`
*   'b' → `1`
*   'c' → `00`
*   'd' → `11`

must be lossy because binary like `0011` could decode to any of 'aabb', 'aad', 'cbb', or 'cd'.

What lossless compression scheme minimizes the expected number of bits per symbol?

Example: The lossless encoding

*   'a' → `0`
*   'b' → `10`
*   'c' → `110`
*   'd' → `1110`

has `15% * 1 + 30% * 2 + 25% * 3 + 30% * 4 = 2.7` expected bits per symbol.  But there is an encoding with better compression.

--------------------------------------------------------------------------------

# Guaranteeing a Lossless Encoding

For a lossless code that can be decoded from left to right without lookahead, the code words for two distinct symbols always take this form:

*   `[common prefix]0[first suffix]`
*   `[common prefix]1[second suffix]`

(Proof: The only other possiblity is that one code word is a prefix of the other.  But after reading the prefix, a no-lookahead decoder would not know whether or not to keep reading bits, which contracts the losslessness of the encoding.)

So if we already have suffixes for two symbols `x` and `y`, we just need to find a common prefix meaning "`x` or `y`", and then we will have the symbols' full code words.

## Example

        Symbol Partitioning      'a'  'b'  'c'  'd'
    --------------------------   ---  ---  ---  ---
    {'a'}, {'b'}, {'c'}, {'d'}
    {'a', 'b'}, {'c'}, {'d'}       0    1
    {'a', 'b', 'c'}, {'d'}        00   01    1
    {'a', 'b', 'c', 'd'}         000  001   01    1

There's still the question of how to choose which sets to combine in each step.  But one clue is that each combination costs a bit for each symbol in the combined set.

--------------------------------------------------------------------------------

# Greedy Algorithm for the Encoding Problem

Problem: Given frequencies for a set of symbols, create a lossless binary encoding that minimizes the expected number of bits per symbols.

## DAG

*   Vertices (situations):
    *   …
*   Edges (actions):
    *   …
*   Edge weights (optional):
    *   …
*   Topological order:
    *   …
*   Goal:
    *   …

## Choosing a Forward Edge

*   Direct solution:
    *   …
    *   …
    *   …

## Example

*   'a' occurs about 15% of the time in a typical text.
*   'b' occurs about 30% of the time in a typical text.
*   'c' occurs about 25% of the time in a typical text.
*   'd' occurs about 30% of the time in a typical text.

             Symbol Partitioning (Frequency Order)           'a'  'b'  'c'  'd'
    -------------------------------------------------------  ---  ---  ---  ---
    {'a'} → 0.15, {'c'} → 0.25, {'b'} → 0.30, {'d'} → 0.30
    …
